@@ -28,8 +28,9 @@ import (
 func main() {
 	fs := modtimevfs.New(ui.Assets, time.Unix(1, 0))
 	err := vfsgen.Generate(fs, vfsgen.Options{
-		PackageName:  "ui",
-		BuildTags:    "!dev",
+		PackageName: "ui",
+		// NOTE(spasquie): using ondiskassets tag becaues Go 1.12 doesn't support comma-separated tags.
+		BuildTags:    "!ondiskassets",
 		VariableName: "Assets",
 	})
 	if err != nil {
